/*
* DL3 Quantization
* ================
*
* File: dl3quant.c
* Copyright (C) 1993-1997 Dennis Lee
*/



#include <cmath>


typedef struct {
    unsigned long r, g, b;
    unsigned long pixel_count;
    float err;
    unsigned long cc;
    unsigned char rr, gg, bb,al;
} CUBE3;

static CUBE3 *rgb_table3;

static unsigned long tot_colors;

//static unsigned long sqr_tbl[255 + 256], *squares3;

static float calc_err(unsigned long c1, unsigned long c2)
{
    unsigned long dist1, dist2, P1, P2, P3;
    long R1, G1, B1, R2, G2, B2, R3, G3, B3;

    P1 = rgb_table3[c1].pixel_count;
    P2 = rgb_table3[c2].pixel_count;
    P3 = P1 + P2;

    R3 = (rgb_table3[c1].r + rgb_table3[c2].r + (P3 >> 1)) / P3;
    G3 = (rgb_table3[c1].g + rgb_table3[c2].g + (P3 >> 1)) / P3;
    B3 = (rgb_table3[c1].b + rgb_table3[c2].b + (P3 >> 1)) / P3;


    R1 = rgb_table3[c1].rr-R3;
    G1 = rgb_table3[c1].gg-G3;
    B1 = rgb_table3[c1].bb-B3;

    R2 = rgb_table3[c2].rr-R3;
    G2 = rgb_table3[c2].gg-G3;
    B2 = rgb_table3[c2].bb-B3;


    R1 *= R1; G1 *= G1; B1 *= B1;
    R2 *= R2; G2 *= G2; B2 *= B2;

    dist1 = R1+G1+B1;
    dist2 = R2+G2+B2;
    
    return (sqrtf((float)dist1) * (float)P1 + sqrtf((float)dist2) * (float)P2);
}

static void recount_next(unsigned long i)
{
    unsigned long j,c2 = 0;
    float err, cur_err;
    err = 1e31f;
    for (j = i + 1; j < tot_colors; j++) {
        cur_err = calc_err(i, j);
        if (cur_err < err) {
            err = cur_err;
            c2 = j;
        }
    }
    rgb_table3[i].err = err;
    rgb_table3[i].cc = c2;
}

static void recount_dist(unsigned long c1)
{
    unsigned long i;
    float cur_err;

    recount_next(c1);
    for (i = 0; i < c1; i++) {
        if (rgb_table3[i].cc == c1) recount_next(i);
        else {
            cur_err = calc_err(i, c1);
            if (cur_err < rgb_table3[i].err) {
                rgb_table3[i].err = cur_err;
                rgb_table3[i].cc = c1;
            }
        }
    }
}

static void reduce_table3()
{
    unsigned long i, c1, c2;
    float err;

    //Quantize Pass 1
    for (i = tot_colors-1; i--; )recount_next(i);

    rgb_table3[tot_colors - 1].err = 1e31f;
    rgb_table3[tot_colors - 1].cc = tot_colors;

    //Quantize Pass 2
    while (tot_colors > 256) {
        err = 1e31f;
        for (i = 0; i < tot_colors; i++) {
            if (rgb_table3[i].err < err) {
                err = rgb_table3[i].err;
                c1 = i;
            }
        }
        c2 = rgb_table3[c1].cc;
        rgb_table3[c2].r += rgb_table3[c1].r;
        rgb_table3[c2].g += rgb_table3[c1].g;
        rgb_table3[c2].b += rgb_table3[c1].b;
        rgb_table3[c2].pixel_count += rgb_table3[c1].pixel_count;

        unsigned long v = rgb_table3[c2].pixel_count, v2 = v >> 1;
        rgb_table3[c2].rr = (unsigned char)((rgb_table3[c2].r + v2) / v);
        rgb_table3[c2].gg = (unsigned char)((rgb_table3[c2].g + v2) / v);
        rgb_table3[c2].bb = (unsigned char)((rgb_table3[c2].b + v2) / v);

        tot_colors--;

        rgb_table3[c1] = rgb_table3[tot_colors];
        rgb_table3[tot_colors - 1].err = 1e31f;
        rgb_table3[tot_colors - 1].cc = tot_colors;

        for (i = 0; i < c1; i++) {
            if (rgb_table3[i].cc == tot_colors) rgb_table3[i].cc = c1;
        }

        for (i = c1 + 1; i < tot_colors; i++) {
            if (rgb_table3[i].cc == tot_colors) recount_next(i);
        }

        recount_dist(c1);
        if (c2 != tot_colors) recount_dist(c2);
    }
}


unsigned long dl3quantfast(unsigned char *inbuf, unsigned long pixels, unsigned char* userpal, unsigned long colorDepth)
{

    rgb_table3 = (CUBE3 *)calloc(sizeof(CUBE3), 32768);
    if (rgb_table3 == NULL) return 0;
    unsigned long i, index;

    for (i = 0; i < pixels; i++) {

        index = ((inbuf[0] & 248) << 7) + ((inbuf[1] & 248) << 2) + (inbuf[2] >> 3);
        rgb_table3[index].r += inbuf[0];
        rgb_table3[index].g += inbuf[1];
        rgb_table3[index].b += inbuf[2];
        rgb_table3[index].pixel_count++;
        inbuf += colorDepth;
    }

    tot_colors = 0;

    for (i = 0; i < 32768; i++)
        if (rgb_table3[i].pixel_count) {
            int v = rgb_table3[i].pixel_count, v2 = v >> 1;
            rgb_table3[i].rr = (unsigned char)((rgb_table3[i].r + v2) / v);
            rgb_table3[i].gg = (unsigned char)((rgb_table3[i].g + v2) / v);
            rgb_table3[i].bb = (unsigned char)((rgb_table3[i].b + v2) / v);
            rgb_table3[tot_colors++] = rgb_table3[i];
        }



    reduce_table3();

    for (i = 0; i < tot_colors; i++) {
        *userpal++ = rgb_table3[i].rr;
        *userpal++ = rgb_table3[i].gg;
        *userpal++ = rgb_table3[i].bb;
        *userpal++ = 0xff;
    }
    free(rgb_table3);
    return(tot_colors);
}
unsigned long dl3quantslow(unsigned char *inbuf, unsigned long pixels, unsigned char* userpal, unsigned long colorDepth)
{
    rgb_table3 = (CUBE3 *)calloc(sizeof(CUBE3), 262144);

    if (rgb_table3 == NULL) return 0;
    unsigned long i, index;

    for (i = 0; i < pixels; i++) {
        index = ((inbuf[0] & 252) << 10) + ((inbuf[1] & 252) << 4) + (inbuf[2] >> 2);

        rgb_table3[index].r += inbuf[0];
        rgb_table3[index].g += inbuf[1];
        rgb_table3[index].b += inbuf[2];
        rgb_table3[index].pixel_count++;
        inbuf += colorDepth;
    }

    tot_colors = 0;
    for (i = 0; i < 262144; i++)

        if (rgb_table3[i].pixel_count) {
            int v = rgb_table3[i].pixel_count, v2 = v >> 1;
            rgb_table3[i].rr = (unsigned char)((rgb_table3[i].r + v2) / v);
            rgb_table3[i].gg = (unsigned char)((rgb_table3[i].g + v2) / v);
            rgb_table3[i].bb = (unsigned char)((rgb_table3[i].b + v2) / v);
            rgb_table3[tot_colors++] = rgb_table3[i];
        }
    reduce_table3();

    for (i = 0; i < tot_colors; i++) {
        *userpal++ = rgb_table3[i].rr;
        *userpal++ = rgb_table3[i].gg;
        *userpal++ = rgb_table3[i].bb;
        *userpal++ = 0xff;
    }
    free(rgb_table3);
    return(tot_colors);
}


